/*
Lgroup -- View and manipulate local groups on workstations, Servers, domains
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  
	cl search32.c /link netapi32.lib odbc32.lib odbccp32.lib mpr.lib version.lib

*/
#include <windows.h>
#include <stdio.h>
#include <time.h> 
#include <lm.h>
#include <io.h>
#include <string.h>

 
#define RTN_OK 0
#define RTN_USAGE 4
#define RTN_NEVER 1
#define RTN_ERROR 8
#define GBUFSIZ 16384
#define USER_COUNT 1024
#define FULLNAMELEN 38

void DisplayErrorText(DWORD dwLastError);
FILE *otfi;
char ErrorMessage[1024], verbose = 0, fullname = 0, FullName[255];
#define MAXSPLIT 32
char Fname[MAXSPLIT][FULLNAMELEN+1];
char slow_but_steady = 0;
USER_INFO_2 *ui2;


/* 
Output Display Formats:

Note:  Group Comments and/or Full Names are only displayed when the -f flag is set

A domainname is either a server name if preceded by \\ or a real domain name

For special accounts, the fullname may display as:			
	[Local Account],[Machine Account],[Domain Server Account],
	[Permit Trust Account] or [Unknown Account Type]

	Verbose Display:
	Delimited:	domainname\groupname, MemberType, domainname\membername, FullName
				Member Types are: 
				GroupName:	first record for each group and consists only of
							domainname\groupname, GroupName,,Group Comment
				Group:		The member is a global group with name domain\groupname
				User:		The member is either a domain user with name domain\username
							or a local user with name \\servername\username
				Unknown:	Don't know what kind of member, domain and username will
							be filled in with whatever is returned from the OS call
				EmptyGroup:	Group has no members

    Non Verbose Display:
	Delimited:	servername, groupname, member type, memberdomain, membergroup, membername, memberfullname
				Member Types are:
				GroupName:	first record for each group and consists only of
							servername, groupname,GroupName,,,,Group Comment
				Group:		If the group is a local group, the member group name
							is empty.  If it is a domain group and we are listing
							global group members, then both domain and name have
							data
				User:		If the group is a local user, then the group name is empty


User Name: (20)
Domain Name: (15)
Group Name: (38)

Default Display:

User Name:           Full Name:               Domain Name:      Group Name:
xxxxxxxxxxxxxxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxx xxxxxxxxxxxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx 

*Local Group:        Group Name:              	                Comment
*MemberGroup:        Group Name:                                Comment  
xxxxxxxxxxxxxxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxx xxxxxxxxxxxxxxxxx xxx...



Verbose Display:
*Local DomainName\GroupName (38)              Comment
*Local xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx xxx...
Group: DomainName\GroupName (38)              Comment
Group: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx xxx... 

If fullname
User:  DomainName\UserName (38)               Full Name: (38)          from group DomainName\GroupName
User:  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxx from group xxx...

else
User:  DomainName\UserName (38)               from group DomainName\GroupName(38)
User:  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx from group xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
*/


//convert sid to a string
BOOL GetTextualSid(
    PSID pSid,            // binary SID
    LPTSTR TextualSid,    // buffer for Textual representation of SID
    LPDWORD lpdwBufferLen // required/provided TextualSid buffersize
    )
{
    PSID_IDENTIFIER_AUTHORITY psia;
    DWORD dwSubAuthorities;
    DWORD dwSidRev=SID_REVISION;
    DWORD dwCounter;
    DWORD dwSidSize;

    // Validate the binary SID.

    if(!IsValidSid(pSid)) return FALSE;

    // Get the identifier authority value from the SID.

    psia = GetSidIdentifierAuthority(pSid);

    // Get the number of subauthorities in the SID.

    dwSubAuthorities = *GetSidSubAuthorityCount(pSid);

    // Compute the buffer length.
    // S-SID_REVISION- + IdentifierAuthority- + subauthorities- + NULL

    dwSidSize=(15 + 12 + (12 * dwSubAuthorities) + 1) * sizeof(TCHAR);

    // Check input buffer length.
    // If too small, indicate the proper size and set last error.

    if (*lpdwBufferLen < dwSidSize)
    {
        *lpdwBufferLen = dwSidSize;
        SetLastError(ERROR_INSUFFICIENT_BUFFER);
        return FALSE;
    }

    // Add 'S' prefix and revision number to the string.

    dwSidSize=wsprintf(TextualSid, TEXT("S-%lu-"), dwSidRev );

    // Add SID identifier authority to the string.

    if ( (psia->Value[0] != 0) || (psia->Value[1] != 0) )
    {
        dwSidSize+=wsprintf(TextualSid + lstrlen(TextualSid),
                    TEXT("0x%02hx%02hx%02hx%02hx%02hx%02hx"),
                    (USHORT)psia->Value[0],
                    (USHORT)psia->Value[1],
                    (USHORT)psia->Value[2],
                    (USHORT)psia->Value[3],
                    (USHORT)psia->Value[4],
                    (USHORT)psia->Value[5]);
    }
    else
    {
        dwSidSize+=wsprintf(TextualSid + lstrlen(TextualSid),
                    TEXT("%lu"),
                    (ULONG)(psia->Value[5]      )   +
                    (ULONG)(psia->Value[4] <<  8)   +
                    (ULONG)(psia->Value[3] << 16)   +
                    (ULONG)(psia->Value[2] << 24)   );
    }

    // Add SID subauthorities to the string.
    //
    for (dwCounter=0 ; dwCounter < dwSubAuthorities ; dwCounter++)
    {
        dwSidSize+=wsprintf(TextualSid + dwSidSize, TEXT("-%lu"),
                    *GetSidSubAuthority(pSid, dwCounter) );
    }

    return TRUE;
}


//
// Unicode entry point and argv
//
int	delimPrint(char *ServerName,char *GpNm,char *GpType,char *DomainName,char *gpname,
			   char *UsrNm,char *FullName,char groupmembers)
{
	char vGpNm[320], vMbrNm[320], vMbrGp[320];
//	static char firsttime=0;

	// Force Domain Name to be server name if empty
	if (strcmp(DomainName,"\\")) {
		strcpy(vMbrGp,DomainName);
		strcpy(vMbrNm,DomainName);
	}
	else {
		strcpy(vMbrNm,ServerName);
		strcpy(vMbrGp,ServerName);
	}
	if (verbose) {
		strcpy(vGpNm,ServerName);
		strcat(vGpNm,"\\");
		strcat(vGpNm,GpNm);
		if (*UsrNm) {
			if (*UsrNm == '\\')
				strcat(vMbrNm,UsrNm);
			else {
				if (strchr(UsrNm,'\\') == NULL) {
					if (vMbrNm[strlen(vMbrNm)-1] != '\\')
						strcat(vMbrNm,"\\");
					strcat(vMbrNm,UsrNm);
				}
				else
					strcpy(vMbrNm,UsrNm);
			}
		}
		else
			*vMbrNm = 0;
		if (*gpname) {
			if (*gpname == '\\')
				strcat(vMbrGp,gpname);
			else {
				if (vMbrGp[strlen(vMbrGp)-1] != '\\')
					strcat(vMbrGp,"\\");
				if (strchr(gpname,'\\'))
					strcpy(vMbrGp,gpname);
				else
					strcat(vMbrGp,gpname);
			}
		}
		else
			*vMbrGp = 0;
		fprintf(otfi,"%s,%s,%s,\"%s\"",vGpNm,GpType,vMbrNm,FullName);
		if (!groupmembers)
			fprintf(otfi,",%s\n",vMbrGp);
		else
			fprintf(otfi,"\n");
	}
	else {
		fprintf(otfi,"%s,%s,%s,%s,%s,%s,\"%s\"\n",ServerName,GpNm,GpType,vMbrNm,gpname,UsrNm,FullName);
	}
	return 0;
}


int GetFullName(LPWSTR wComputerName, LPWSTR wUserName, char *FullName, int fullnameSize) {
	DWORD nas;

	nas = NetUserGetInfo(wComputerName,wUserName,2,(LPBYTE *) &ui2);
	if (nas == NERR_Success) {
		if (ui2->usri2_flags & UF_NORMAL_ACCOUNT) {
			sprintf(FullName,"%ls",ui2->usri2_full_name);
//			if (strlen(FullName) > fullnameSize) 
//				strcpy(FullName+20," ...");
//			strcat(UserName,FullName);
		}
		else if (ui2->usri2_flags & UF_TEMP_DUPLICATE_ACCOUNT)
			strcpy(FullName,"[Local Account]");
		else if (ui2->usri2_flags & UF_WORKSTATION_TRUST_ACCOUNT)
			strcpy(FullName,"[Machine Account]");
		else if (ui2->usri2_flags & UF_SERVER_TRUST_ACCOUNT)
			strcpy(FullName,"[Domain Server Account]");
		else if (ui2->usri2_flags & UF_INTERDOMAIN_TRUST_ACCOUNT)
			strcpy(FullName,"[Permit Trust Account]");
		else 
			strcpy(FullName,"[Unknown Account Type]");
		NetApiBufferFree((PVOID) ui2);
	}
	else  {
		fprintf(stderr,"Error retrieving user information for: %ls\n", wUserName);	
		if (otfi != stdout)
			fprintf(otfi,"%-38ls Error Retrieving user information.\n",wUserName);
		DisplayErrorText( nas );
		return RTN_ERROR;
	}
	return 0;
}


void add_new_domain(char **Domn, LPWSTR *wPDC, int *DomainKt, int *DomainMax, 
					LPWSTR wDomNm, LPWSTR wPDCname) {

	int l;
	char PDC[32];
	// Increase size of arrays if necessary
	if (*DomainKt >= *DomainMax) {
		*DomainMax += 64;
		wPDC = realloc(wPDC, sizeof(LPWSTR) * (*DomainMax));
		Domn= realloc(Domn, sizeof(char *) * (*DomainMax));
	}

	// Add name to list as wchar and char
	l = wcslen(wDomNm);
	sprintf(PDC,"%ls",wDomNm);
	Domn[*DomainKt] = malloc(l+1);
	l = wcslen(wPDCname);
	sprintf(PDC,"%ls",wPDCname);
	wPDC[*DomainKt] = malloc(sizeof(WCHAR) * (l + 1));
	sprintf(Domn[*DomainKt],"%ls",wDomNm);
	wcscpy(wPDC[*DomainKt],wPDCname);
	sprintf(PDC,"%ls",wPDC[*DomainKt]);
	++(*DomainKt);
}

// Break the field into pieces of apx the same length
int splitname(char *FullName,int fullnamelen) {
int ln, ii, jj, offset, charstocopy;
char temp[256];

ln = strlen(FullName);
Fname[0][0] = 0;
temp[0] = 0;
if (ln <= fullnamelen) {
	strcpy(Fname[0],FullName);
	return 1;
}

// if not delim and name is longer than fullnamelen then split it into multiple pieces
	ii = 0;
	offset = 0;
	while (offset < strlen(FullName) && ii < MAXSPLIT) {
		// Calculate how many characters to move
		charstocopy = strlen(FullName) - offset;
		if (charstocopy > fullnamelen)
			charstocopy = fullnamelen;
		strcpy(temp,FullName+offset);
		jj = charstocopy;
		// look for 1st comma on the rights
		if (strlen(temp) >  fullnamelen) {
			while (jj >0 && temp[jj] != ',')
				--jj;
		}
		// if found the delimiter, split
		if (jj > 0) {
			offset+=jj+1;
			temp[jj+1] = 0;
		}
		else {
			// no delimiter, split at fixed size
			temp[charstocopy] = 0;
			offset += charstocopy;
		}
	
		strcpy(Fname[ii],temp);
		++ii;
	}
return ii;
}


int ListGlobalGroupMembers(LPWSTR gpname, int *DomainKt, int *DomainMax,
	 char **Domn, LPWSTR *wPDC, FILE *otfi, char dlim,char *GpNm, 
	 char *ServerName, char groupmembers) {
// Parse out domain name and group name; list all members of the group
       LPWSTR		wdelim;
	   char			*delim, gpnm[255];
       LPWSTR       wComputerName; // default to local machine
	   WCHAR		wGroupName[64];

	   GROUP_USERS_INFO_0 *gi0;
	   GROUP_INFO_1 *groupinfo1;
	   DWORD entries_read, total_entries, resume_handle;

   char UserName[255],  ComputerName[32], DomainName[64];
   char GroupName[257], groups[320];
	
   unsigned i,localgrouplen=GBUFSIZ,os, delimpos;
   int dom, fullnamelen = FULLNAMELEN, ii=0, jj=0;

   NET_API_STATUS  nas;
   sprintf(groups, "%ls", gpname);

	wdelim = wcschr(gpname, L'\\');
	delimpos = wdelim - gpname;

	if (wdelim == NULL) {
		fprintf(stderr,"%ls -- Invalid Group Name\n", gpname);
		return 4;
	}
	*wdelim = 0;
	delim = strchr(groups, '\\');
	*delim = 0;
	wcscpy(wGroupName, wdelim+1);
//	wcscpy(wComputerName, gpname);
	sprintf(ComputerName,"%ls", gpname);
	sprintf(GroupName,"%ls",wGroupName);
	sprintf(DomainName,"%ls", gpname);

	if (gpname[0] != L'\\' || gpname[1] != L'\\') {
		for (dom=0,wComputerName=NULL; dom < *DomainKt && wComputerName==NULL; ++dom) {
			if (stricmp(Domn[dom],groups) == 0)
				wComputerName = wPDC[dom];
		}
		if (wComputerName == NULL) {
			// PDCName not found; look up on network and add to list
			nas = NetGetDCName(NULL,gpname,(LPBYTE *)&wComputerName);
			if (nas == NERR_Success) {
				add_new_domain(Domn, wPDC, DomainKt, DomainMax, gpname, wComputerName);
				*delim = L'\\';
//				fprintf(stderr,"Located PDC for domain: %s\n",Domn[DomainKt]);
				NetApiBufferFree((PVOID) wComputerName);
				wComputerName=wPDC[*DomainKt - 1];
				sprintf(ComputerName,"%ls",wComputerName);
			}	
			else {
				fprintf(stderr,"Unable to locate domain controller for %ls\n", gpname);
				DisplayErrorText( nas );
				if (otfi != stdout)
					fprintf(otfi,"%s\n",ErrorMessage);
				if (otfi == stdout)
					return RTN_ERROR;
				fclose(otfi);
				wComputerName = NULL;
				return RTN_ERROR;
			}
		}
	}
	else
		nas = NERR_Success;


	*wdelim = L'\\';		// make gpname a single variable again for printing
    sprintf(gpnm,"%ls",gpname);
	sprintf(ComputerName,"%ls",wComputerName);
	strupr(ComputerName);
    // Get the type 0 info into pgm2

	entries_read = 0;
	total_entries= 1;
	*groups = 0;
	resume_handle = 0;

	if (fullname) {
		// Get the group Comment by doing a NetGroupGetInfo
		// Print it out;
		nas=NetGroupGetInfo(wComputerName,wGroupName,1,(LPBYTE *)&groupinfo1);
		if (nas==NERR_Success) {
			if (!dlim && *groupinfo1->grpi1_comment)
				fprintf(otfi," %ls\n",groupinfo1->grpi1_comment);
			NetApiBufferFree(groupinfo1);
		}
		else  {
			fprintf(stderr,"Error retrieving group information for Group: %ls in %ls\n", 
				wGroupName, wComputerName);
			DisplayErrorText( nas );
			if (dlim)
				delimPrint(ServerName,GpNm,"Error",DomainName,ComputerName,GroupName,ErrorMessage,0);
			else
				fprintf(otfi,"Group:   %ls\\%ls. %s\n", wComputerName, wGroupName, ErrorMessage);
			return RTN_ERROR;
		}
	}

	// Read in all of the users
	while (entries_read < total_entries) {
		nas = NetGroupGetUsers(wComputerName,wGroupName,0,(LPBYTE *) &gi0,
			GBUFSIZ,&entries_read, &total_entries, &resume_handle);

		if (nas == NERR_Success || nas == ERROR_MORE_DATA) {
			// Found More
			for (i=0; i < entries_read; ++i) {
				if (verbose && !dlim)
					sprintf(UserName,"%s\\%ls", strupr(DomainName), gi0[i].grui0_name);
				else
					sprintf(UserName,"%ls", gi0[i].grui0_name);
				
				*FullName = 0;
				if (fullname) {
					if (GetFullName(wComputerName, gi0[i].grui0_name, FullName,dlim?254:fullnamelen) == RTN_ERROR)
						return RTN_ERROR;
					// attempt to split at commas if possible
					if (!dlim) 
						splitname(FullName,FULLNAMELEN);			
				}

				if (dlim) {
					delimPrint(ServerName,GpNm,"GroupUser",DomainName,gpnm,UserName,FullName,groupmembers);
					continue;
				}
				
				if (verbose) {
					if (fullname) {
						fprintf(otfi,"User:  %-38s %-38s in Group %ls\n",UserName,Fname[0],gpname);
						for (jj=1;jj < ii;++jj) 
							fprintf(otfi,"       %-38s %-38s\n"," ",Fname[jj]);
					}
					else
						fprintf(otfi,"User:  %-38s in Group %ls\n",UserName,gpname);
					continue;
				}
				
				if (fullname) {
					fprintf(otfi,"%-20s %-38s %-17s %ls\n", UserName, Fname[0], DomainName, gpname);
					for (jj=1;jj < ii;++jj)
						fprintf(otfi,"%-20s %-38s\n", " ", Fname[jj]);
				}
				else
					fprintf(otfi,"%-59s %-17s %ls\n", UserName, strupr(DomainName), gpname);		
			}
			
			if (entries_read == 0)
				if (dlim)
					delimPrint(ServerName,GpNm,"EmptyGroup",DomainName,gpnm,"","",groupmembers);
		}
		else  {
			fprintf(stderr,"Error retrieving group users for Group: %ls in %ls\n", 
				wGroupName, wComputerName);
			DisplayErrorText( nas );
			if (dlim) {	
				delimPrint(ServerName,GpNm,"Error",ComputerName,GroupName,"",ErrorMessage,0);
			}
			else
				fprintf(otfi,"Group:   %ls\\%ls. %s\n", wComputerName, wGroupName, ErrorMessage);
			return RTN_ERROR;
		}
	}	
	
	NetApiBufferFree((PVOID) gi0);
	return 0;

}

void help( wchar_t *argv[],char *tmp) {
		// Use Notepad to display help
	FILE *hlp;
	char cmd[1024];
//	char compiledate[] = "__DATE__";
	strcat(tmp,"lgroup.hlp");
	hlp = fopen(tmp,"w");

	
	fprintf(hlp,"View / Manipulate Local Groups Copyright London Computing, 1985-2004, (WIN32) %s.\n",__DATE__);
	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
	fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(hlp,"(at your option) any later version.\n\n");
	fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
	fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(hlp,"GNU General Public License for more details.\n\n");
	fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
	fprintf(hlp,"along with this program; if not, write to the Free Software\n");
	fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
	
	fprintf(hlp,"LGROUP Ver 1.26 %s\n",__DATE__);
	fprintf(hlp, "Usage:\n%ls {<group> | -List} {<servername>|<domainname> ...} [-M] [-V] [-T] [-F] [-G] [-D]\n",argv[0]);
	fprintf(hlp, "Displays all of the members of the specified local group.\n\n");
	fprintf(hlp, "Usage:\n%ls <group>  {<servername>|<domainname> ...} {[-Addgroup] [-Comment:\"comment\"] | -Delgroup }\n",argv[0]);
	fprintf(hlp, "Adds / Deltes groups.\n\n");
	fprintf(hlp, "Usage:\n%ls {<group> | -List} {<servername>|<domainname> ...}  [-1] {-Add | -Del | -Set} mbr ... \n",argv[0]);
	fprintf(hlp, "Adds members to the group, removes members from the group or sets the group membership.\n\n");
	fprintf(hlp, "<group>      is the local group name to Check\n");
	fprintf(hlp, "-List		  if specified, return a list of all groups.\n");
	fprintf(hlp, "             Both Group Name and Group Comment are displayed.\n");
	fprintf(hlp, "             If -T specified, only Group Names will be displayed.\n");
	fprintf(hlp, "<servername> is the name of a Domain Controller (W2K or above) or NT PDC or NT BDC.\n");
	fprintf(hlp, "             The server name must begin with \\\\\n");
	fprintf(hlp, "<domainname> is the name of an NT or W2K domain, must NOT begin with \\\\\n");
	fprintf(hlp, "             You may specify multiple servers or domains.  If the\n");
	fprintf(hlp, "             name begins with an @, it is assumed to be the name\n");
	fprintf(hlp, "             of a file containing a list of servers, one per line. The @\n");
	fprintf(hlp, "             is not part of the file name.\n\n");
	fprintf(hlp, "-Addgroup    Adds the group to the server or domain.  If -Comments specified.\n");
	fprintf(hlp, "             Adds the comment as well.\n\n");
	fprintf(hlp, "-Comment:cc  cc is a comment to be added to the group when it is created.\n");
	fprintf(hlp, "             cc should be enclosed in quotes (\").  If -Addgroup not specified,\n");
	fprintf(hlp, "             the group comment will be changed.\n\n");
	fprintf(hlp, "-Delgroup    Removes the group from te server of domain.\n\n");
	fprintf(hlp, "-Add         Adds the following members to the specified group.\n\n");
	fprintf(hlp, "-Del         Deletes the following members from the specified group.\n\n");
	fprintf(hlp, "-Set         Sets the group to contain only the following members.\n\n");
	fprintf(hlp, " Mbr         A member name can either be the name of a user\n");
	fprintf(hlp, "             or a global group.  Member names can be entered either\n");
	fprintf(hlp, "             with or without a preceding domain name name.  If no domain\n");
	fprintf(hlp, "             is specified, the current domain or server will be used. If the\n");
	fprintf(hlp, "             User or group name contains whitespace characters, it must\n");
	fprintf(hlp, "             be enclosed in quotes.  Examples:\n");
	fprintf(hlp, "             User1 \"Group A\" \"Southeast\\Group B\" northeast\\user2\n");
	fprintf(hlp, "             If the Mbr begins with a @, it is assumed to be the name\n");
	fprintf(hlp, "             of a file containing a list of members, one per line. The @\n");
	fprintf(hlp, "             is not part of the file name.  In this case, quotes should\n");
	fprintf(hlp, "             should NOT be used.\n\n");
	fprintf(hlp, "-F           Display Full Name as well as User Name\n\n");
	fprintf(hlp, "             Display group comment for each group.\n");
	fprintf(hlp, "-G           Ignored unless -List specified.  If specified, lists local group\n");
    fprintf(hlp, "             members including users and global groups, but does not list\n");
	fprintf(hlp, "             the members of the Global Groups.  Contrast with -M which lists\n");
	fprintf(hlp, "             global group members as well.\n\n");
	fprintf(hlp, "-M           Ignored unless -List specified.  If specified, lists members\n");
	fprintf(hlp, "             for each defined local group.\n\n");
	fprintf(hlp, "-T           Terse mode.  Do not display users in any global groups\n\n");
	fprintf(hlp, "             defined as part of this local group.\n");
	   fprintf(hlp, "-1           Used only with -Add or -Del.  Makes all changes one at a time.\n");
	   fprintf(hlp, "             instead of attempting to add or delete the entire group with one\n");
	   fprintf(hlp, "             call.  This takes more time, but is more reliable.\n\n");
	   fprintf(hlp, "-D           Used when listing members only. Output data in a comma delimited format\n");
	   fprintf(hlp, "             one entry per line as:\n\n");
	   fprintf(hlp, "             servername, localgroupname, entrytype, domainname, groupname, username, fullname.\n");
	   fprintf(hlp, "             entrytype is: User or Global or Universal.  When -G specified, either\n");
	   fprintf(hlp, "             username or groupname will be empty since global group members are not.\n");
	   fprintf(hlp, "             displayed.  \n\n");
	   fprintf(hlp, "             When -V is specified as well, each line looks like:\n\n");
	   fprintf(hlp, "             severname\\localgroupname, entrytype, domain\\groupname, domain\\username, fullname.\n\n");
	   fprintf(hlp, "-V           Prefix user and group names with source.  See below for an explanation.\n\n");
	   fprintf(hlp, "Local Group membership includes both explicit membership (the user was assigned\n");
	   fprintf(hlp, "to a local group) and implicit membership (the user is a member of the local \n");
	   fprintf(hlp, "group because he/she is a member of a global group that is in turn a member of\n");
	   fprintf(hlp, "local group), e.g. User is a member of the local group Administrators because\n");
	   fprintf(hlp, "the user is a member of domain admins which is in turn a member of the local\n");
	   fprintf(hlp, "administrators group. Skip implicit membership if -T or -G is specified.\n\n");
	   fprintf(hlp, "The output of this command can be displayed in one of two formats depending on\n");
	   fprintf(hlp, "presence or absence of the -V switch.  When -V is present, the output format\n");
	   fprintf(hlp, "differs, with more explanation of the data being displayed.  The first line\n");
	   fprintf(hlp, "of the file is always the name of the Local Group.  When -V is set every line\n");
	   fprintf(hlp, "of the output file is written in one of three formats:\n\n");
	   fprintf(hlp, "User:    <domain name>\\<user name> [fullname] from group <domain name>\\<group name>]\n");
	   fprintf(hlp, "Group:   <domain name>\\<group name> [comment]\n");
	   fprintf(hlp, "Deleted: <domain name>\\<user name>\n\n");
	   fprintf(hlp, "A User line is written for each user explicitly defined to the local group.\n");
	   fprintf(hlp, "The user information is specified as <domain name>\\<user name>.  In this\n");
	   fprintf(hlp, "case, there is no group information.  A Group line is written for every \n");
	   fprintf(hlp, "global group defined in the local group.  The group information is displayed\n");
	   fprintf(hlp, "as <domain name>\\<group name>.  Following the Group line are a series of user\n");
	   fprintf(hlp, "lines that list all of the users in the global group.  These user lines do\n");
	   fprintf(hlp, "include group information.  The example below shows group ABC in domain TEST\n");
	   fprintf(hlp, "which has three members: User1 in domain test, User2 in domain test1 and Group1\n");
	   fprintf(hlp, "in domain test.  Group1 has two members: User3 and User4.  Example:\n\n");
	   fprintf(hlp, "*Local:  TEST\\ABC          [Comment]\n");
	   fprintf(hlp, "User:    test\\User1        [Full Name]\n  ");
	   fprintf(hlp, "User:    test1\\User2       [Full Name]\n");
	   fprintf(hlp, "Group:   test\\Group1       [Comment]\n");
	   fprintf(hlp, "User:    test\\User3        [Full Name] from group test\\Group1\n");
	   fprintf(hlp, "User:    test\\User4        [Full Name] from group test\\Group2\n");
	   fprintf(hlp, "Full Names and Group Comments are displayed when the -F flag is used.\n");
	   fprintf(hlp, "The full name will be truncated to 38 characters.  Comments are not truncated.\n\n");
	
	   fprintf(hlp, "In the default mode data is displayed as follows:\n");
	   fprintf(hlp, "<user name>  [Full Name]	<domain name>	[group name] -or-\n");
	   fprintf(hlp, "*MemberGroup <Group Name>                  [comment]\n");
	   fprintf(hlp, "depending on whether the group member is a user or a global group.\n");
	   fprintf(hlp, "Example (Same information as previous example.):\n");
	   fprintf(hlp, "*Local Group:  TEST\\ABC                    [Comment]\n");
	   fprintf(hlp, "User1          test       [Full Name]\n");
	   fprintf(hlp, "User2          test1      [Full Name]\n");
	   fprintf(hlp, "*Member Group: test\\Group1                 [Comment]\n");
	   fprintf(hlp, "User3          test       [Full Name]      test\\Group1\n");
	   fprintf(hlp, "User4          test       [Full Name]      test\\Group1\n");
	   fprintf(hlp, "Full Names and Group Comments are displayed when the -F flag is used.\n");
	   fprintf(hlp, "The full name will be truncated to 38 characters.  Comments are not truncated.\n\n");
	   fprintf(hlp, "The output of this command will be displayed in a NOTEPAD window unless it is\n");
	   fprintf(hlp, "redirected to a file using > or >> or piped to another command using |.\n\n");
	fprintf(hlp, "%s%s%s%s",
	             "Returns:     0--Valid Last Logon Displayed\n",
	             "             1--Never Logged on to Server\n",
				 "             4--This help displayed.\n",
				 "             8--Parameter Error.\n");
   	fclose(hlp);
	strcpy(cmd,"NOTEPAD.EXE ");
	strcat(cmd,tmp);

	WinExec(cmd,SW_SHOW);
}


void set_flags(int argc, wchar_t *argv[], char *members, char *verbose, 
			   char *fullname, char *terse, char *groupmembers, 
			   int *starting_member, char *add, char *del, char *set,
			   char *slow_but_steady, char *delim, char *addgroup, 
			   char *delgroup, LPWSTR wComments, char *excludes) {
	int i;
	char Args[255];
	*excludes = 0;

   //
   // process command line arguments
   //
	for (i=3; i < argc && *add==0 && *del==0; ++i) {
		sprintf(Args,"%ls",argv[i]);
		if (stricmp(Args,"-1") == 0)
			*slow_but_steady = 1;
		if (stricmp(Args,"-D") == 0)
			*delim = 1;
		if (stricmp(Args,"-M") == 0)
			*members = 1;
		if (stricmp(Args,"-V") == 0)
			*verbose = 1;
		if (stricmp(Args,"-F") == 0)
			*fullname = 1;
		if (stricmp(Args,"-T") == 0)
			*terse = 1;
		if (stricmp(Args,"-G") == 0) {
			*groupmembers = 1;
			*members = 1;
		}
		if (Args[0]=='-' && toupper(Args[1])=='X') {
			strcat(excludes,",");
			strcat(excludes,Args+2);
		}
		if (stricmp(Args,"-add") == 0) {
			*add = 1;
			*starting_member = i+1;
		}
		if (stricmp(Args,"-Del") == 0) {
			*del = 1;
			*starting_member = i+1;
		}
		if (stricmp(Args,"-SET") == 0) {
			*set = 1;
			*starting_member = i+1;
		}
		if (stricmp(Args,"-ADDGROUP") == 0) {
			*addgroup = 1;
		}
		if (stricmp(Args,"-DELGROUP") == 0) {
			*delgroup = 1;
		}
		if (strncmp(strupr(Args),"-COMMENT:",9) == 0) {
			wcscpy(wComments, argv[i]+9);
		}
		if (strncmp(strupr(Args),"-C:",3) == 0) {
			wcscpy(wComments, argv[i]+3);
		}

	}
}
 

void expand_abbreviation(LPWSTR wDomNm, LPWSTR wComputerName) {
	char ComputerName[32];
	sprintf(ComputerName,"%ls",wComputerName);
	strupr(ComputerName);
	wcscpy(wDomNm,wComputerName);
}


FILE *OpenOutfile(const int stdio_handle, const char *GroupName, const char *ComputerName, 
				  const char list, const char *temppath, char *filepath) {

	FILE *otfi;
	unsigned i;

	strcpy(filepath, temppath);
	if(! _isatty(stdio_handle) )
		// stdout has been redirected, don't create a temp file for notepad
		otfi = stdout;
	else {
		if (list)
			strcat(filepath,"LocalGroups");
		else
			strcat(filepath,GroupName);
		// skip past any leading \\ in the computername
		for (i=0;i < strlen(ComputerName) && ComputerName[i]=='\\';++i);
		if (ComputerName[i] == '-')
			++i;				// Skip past - in -NE...
		strcat(filepath,"_");
		strcat(filepath,ComputerName+i);
		strcat(filepath,".txt");
		otfi = fopen(filepath,"w");
		if (otfi == NULL) {
			fprintf(stderr, "Unable to create %s\n",filepath);
			fprintf(stderr, "Output will be sent to STDOUT\n");
			otfi = stdout;
		}
	}
	return otfi;
}


NET_API_STATUS add_to_domainlist(FILE *otfi, LPWSTR wDomNm, 
	LPWSTR wCN, char **Domn, LPWSTR *wPDC, LPWSTR wComputerName, 
	int *DomainKt, int *DomainMax) {
	DWORD nas;
	// Check to see if the current computername is a domain name 
	// or a computername
	// Add domain name to Domn, PDC name to wPDC, inc 
	char temp[255];
	LPWSTR wServer;
	// wServer = NULL;
	sprintf(temp,"%ls",wDomNm);
	if (wCN[0] != L'\\' || wCN[1] != L'\\') {
		nas = NetGetDCName(NULL,(LPWSTR) wDomNm,(LPBYTE *)&wServer);	
		if (nas == NERR_Success) {
			wcscpy(wComputerName,wServer);
			sprintf(temp,"%ls",wComputerName);
			NetApiBufferFree(wServer);
			add_new_domain(Domn, wPDC, DomainKt, DomainMax, 
				wDomNm, wComputerName);
		}
	}
	else { 
		//Add machine name as its own domain
		//Done in case we look at workstations or member servers
		add_new_domain(Domn, wPDC, DomainKt, DomainMax, wCN+2, wCN);
		nas = NERR_Success;
	}

	if (nas != NERR_Success) {
		fprintf(stderr,"Unable to locate domain controller for %ls\n",
			wComputerName);
		DisplayErrorText( nas );
		if (otfi != stdout)
			fprintf(otfi,"%s\n",ErrorMessage);
		if (otfi == stdout)
			fclose(otfi);
		return RTN_ERROR;
    }
	else
		return NERR_Success;
}

LPWSTR *gather_computer_names(wchar_t *argv[], int starting_member,
						   int *CNameCount, int *CNameArraySize, FILE *otfi) {

	FILE *inlist;
	char arg[_MAX_PATH], temp[_MAX_PATH];
	int k,i;
	LPWSTR *wCName;

	wCName = (LPWSTR *)malloc(*CNameArraySize * sizeof(LPWSTR));
	fprintf(otfi==stdout?stderr:otfi, "The following domains / computers will be queried:\n");
	for (i=2; i < starting_member - 1; ++i) {
		// Look at each entry, should specify a computer or domain name
		// If the entry starts with an @, read the file specified (name w/o @)
		// and add name into wCname array.  Realloc array if too small
		// If the entry does not start with an @, add the name to the array
		sprintf(arg,"%ls",argv[i]);
		if (*arg == '@') {
			inlist = fopen(arg+1,"r");
			if (inlist == NULL) {
				fprintf(stderr, "Unable to open server list: %ls\n", argv[i]+1);
			}
			else {
				while(fgets(temp,sizeof(temp),inlist)) {
					temp[strlen(temp)-1] = 0; // remove ending lf
					wCName[*CNameCount]= malloc(sizeof(WCHAR)*(strlen(temp)+1));
					k=mbstowcs(wCName[*CNameCount],temp,strlen(temp)+1);
					++(*CNameCount);
					if (*CNameCount >= *CNameArraySize) {
						*CNameArraySize += 64;
						wCName = realloc(wCName, *CNameArraySize * sizeof(LPWSTR));
					}
					
					fprintf(otfi==stdout?stderr:otfi,"\t%s\n",temp);
				}
				fclose(inlist);
			}
		}
		else {
			if (*arg == '-')
				// Skip any names starting with a -, they are flags
				continue;
			else {
				// add the name to the wCName array for later processing
				wCName[*CNameCount] = malloc(sizeof(WCHAR)*wcslen(argv[i])+1);
				wcscpy(wCName[*CNameCount], argv[i]);
				sprintf(temp,"%ls",wCName[*CNameCount]);
				++(*CNameCount);
				if (*CNameCount >= *CNameArraySize) {
					*CNameArraySize += 64;
					wCName = realloc(wCName, *CNameArraySize * sizeof(LPWSTR));
				}
				fprintf(otfi==stdout?stderr:otfi,"\t%s\n",temp);
			}
		}
	}
	return wCName;
}


void gather_users(
	FILE *otfi, int starting_member, int argc, wchar_t **argv, 
	LOCALGROUP_MEMBERS_INFO_3 *localgroup_members, int *count) {
	
	int i,k;
	FILE *inlist;
	char temp[_MAX_PATH];

	// save the list of users to add or delete
	for (*count=0,i=starting_member; i < argc && *count < USER_COUNT; ++i) {
		sprintf(temp,"%ls",argv[i]);
		if (temp[0] == '@') {
			inlist = fopen(temp+1,"r");
			if (inlist == NULL) {
				if (otfi != stdout)
					fprintf(otfi,"Unable to open: %s\n",temp+1);
				fprintf(stderr,"Unable to open: %s\n",temp+1);
			}
			else {
				while(fgets(temp,sizeof(temp),inlist) && *count < USER_COUNT) {
					temp[strlen(temp)-1] = 0; // remove ending lf
					localgroup_members[*count].lgrmi3_domainandname = malloc(sizeof(WCHAR)*(strlen(temp)+1));
					k=mbstowcs(localgroup_members[*count].lgrmi3_domainandname,temp,strlen(temp)+1);
				    sprintf(temp,"%ls",localgroup_members[*count].lgrmi3_domainandname);
					++(*count);
					// fprintf(otfi,"\t%s\n",temp);
				}
				fclose(inlist);
			}
		}
		else {
			// localgroup_members[*count].lgrmi3_domainandname = argv[i];
			localgroup_members[*count].lgrmi3_domainandname = malloc(sizeof(WCHAR)*(wcslen(argv[i])+1));
			wcscpy(localgroup_members[*count].lgrmi3_domainandname, argv[i]);
			sprintf(temp,"%ls",localgroup_members[*count].lgrmi3_domainandname);
			++(*count);
		}
	}
}


int add_delete_groups(FILE *otfi, LPWSTR wComputerName, LPWSTR wGroupName, 
	LPWSTR wComments, const char add) {

	int rc;
	NET_API_STATUS  nas;
	char action[32];
	char temp[320];
	LOCALGROUP_INFO_1 g;
	DWORD parm_err;
/*
Value Meaning 
ERROR_ACCESS_DENIED  The user does not have access to the requested information. 
NERR_InvalidComputer The computer name is invalid. 
NERR_NotPrimary The operation is allowed only on the primary domain controller of the domain. 
NERR_GroupExists The local group already exists. 
ERROR_ALIAS_EXISTS The local group already exists. 

*/

	if (add=='A') {
		g.lgrpi1_name = wGroupName;
		g.lgrpi1_comment = wComments;
		strcpy(action,"Add Group");
		nas = NetLocalGroupAdd(wComputerName, 1, (LPBYTE)&g, &parm_err);
	}
	else if (add=='C') {
		g.lgrpi1_name = wGroupName;
		g.lgrpi1_comment = wComments;
		strcpy(action,"New Comment");
		nas = NetLocalGroupSetInfo(wComputerName, wGroupName, 1, (LPBYTE)&g, &parm_err);
	}
	else {
		strcpy(action,"Delete Group");
		nas = NetLocalGroupDel(wComputerName, wGroupName);
	}

	// check the status of the command
	if (nas == NERR_Success) {
		fprintf(otfi,"%s %ls\\%ls: Success\n",action,wComputerName,wGroupName);
		rc = 0;
	}
	else {
		sprintf(temp,"%ls\\%ls",wComputerName,wGroupName);
		DisplayErrorText( nas );
		fprintf(otfi, "%s %s: Failed, Error Number %d.  %s",temp, action, nas, ErrorMessage);
		rc = -1;
	}

	return rc;
}


int add_delete_users(FILE *otfi, LPWSTR wComputerName, LPWSTR wGroupName, 
	 const char add, const char del, const char set, 
	 LOCALGROUP_MEMBERS_INFO_3 *localgroup_members, int count) {

	char temp[_MAX_PATH], member[_MAX_PATH];
	int i, rc;
   NET_API_STATUS  nas;


	// Print out list of users to add / delete
   if (slow_but_steady) 
	   sprintf(temp,"%ls\\%ls %s ",
	   wComputerName, wGroupName,
	   (add ? "Adding" : (del ? "Deleting" : (set ? "Setting" : "???"))));
   else {
   		sprintf(temp, "Local Group %ls\\%ls\n%s Members:\n",
		wComputerName, wGroupName, (add ? "Adding" : (del ? "Deleting" : (set ? "Setting" : "???"))));
		fprintf(otfi,temp);
   }
	
 
	for (i=0; i < count; ++i) {
		sprintf(member,"%ls",localgroup_members[i].lgrmi3_domainandname);
		if (!slow_but_steady) {
			fprintf(otfi,"\t%s\n",member);
			continue;
		}

		if (!add && !del) {
			fprintf(otfi,"%s %s: Skipped.\n",temp,member);
			continue;
		}

		if (add)
			nas = 
			NetLocalGroupAddMembers(wComputerName,wGroupName,3,(LPBYTE)(localgroup_members+i),1); 
		else 
			nas = 
			NetLocalGroupDelMembers(wComputerName,wGroupName,3,(LPBYTE)(localgroup_members+i),1); 

		// check the status of the command
		if (nas == NERR_Success) {
			fprintf(otfi,"%s %s: Success\n",temp,member);
			rc = 0;
		}
		else {
			DisplayErrorText( nas );
			fprintf(otfi, "%s %s: Failed.  %s",temp, member, ErrorMessage);
			rc = -1;
		}

	}

	if (slow_but_steady && (add || del)) 
		// Already done
		return rc;

	// You get here if Not slow_but_steady or a set operation or count was 0
	if (add)
		nas = 
		NetLocalGroupAddMembers(wComputerName,wGroupName,3,(LPBYTE)localgroup_members,count); 
	else if (del)
		nas = 
		NetLocalGroupDelMembers(wComputerName,wGroupName,3,(LPBYTE)localgroup_members,count); 
	else if (set)
		nas = 
		NetLocalGroupSetMembers(wComputerName,wGroupName,3,(LPBYTE)localgroup_members,count); 
	else {
		fprintf(otfi,"Nothing to do.  Add == 0, Del == 0, Set == 0\n");
		rc = -1;
		return rc;
	}


	if (nas == NERR_Success) {
		fprintf(otfi,"Success\n");
		rc = 0;
	}
	else {
		DisplayErrorText( nas );
		fprintf(otfi, "%s\n", ErrorMessage);
		rc = -1;
	}
	return rc;
}


// Enumerate groups or create a list of exactly 1 group name
NET_API_STATUS Enum_Groups(char list, LPWSTR wComputerName, LPWSTR wGroupName,
  LOCALGROUP_INFO_1 **lgi1, LOCALGROUP_INFO_1 *LocalGroupInfo1, DWORD *entriesread, 
  DWORD *totalentries, DWORD *resumehandle,  LPWSTR wComment, FILE *otfi) {
	// If we need to enum groups, call the enum function, otherwise we have a group name
	// pretend that we called enum so the code that displays the data works the same way
	// for either case

	NET_API_STATUS nas;
	DWORD tot = *totalentries;
	DWORD res = *resumehandle;
	DWORD ent = *entriesread;


	if (list) {
		nas = NetLocalGroupEnum(wComputerName,1, (LPBYTE *) lgi1,
			GBUFSIZ, &ent,  &tot, &res );
		*entriesread = ent;
		*totalentries= tot;
		*resumehandle= res;
	}
	else {
		nas = NERR_Success;
		*entriesread=1;
		*totalentries=1;
		LocalGroupInfo1[0].lgrpi1_name = wGroupName;
		*wComment=0;
		LocalGroupInfo1[0].lgrpi1_comment = wComment;
		*lgi1 = LocalGroupInfo1;
	}

		// If call failed, print error message
		if (nas != NERR_Success && nas != ERROR_MORE_DATA) {
			if (*lgi1)
				NetApiBufferFree((PVOID) *lgi1);
			fprintf(stderr,"Unable to enumerate groups for %ls\n", wComputerName);
			DisplayErrorText( nas );
			if (otfi != stdout)
				fprintf(otfi,"%s\n",ErrorMessage);
			if (otfi != stdout)
				fclose(otfi);
			return RTN_ERROR;
		}
		else
			return nas;
}

NET_API_STATUS GetLocalGroupComment(LPWSTR wComputerName, LPWSTR wGroupName, 
									LPWSTR wComment, FILE *otfi, char *OutFileName){
	LOCALGROUP_INFO_1 *lgi2;

	NET_API_STATUS nas;
	char cmd[_MAX_PATH+32];

	nas = NetLocalGroupGetInfo(wComputerName,wGroupName,1, (LPBYTE *) &lgi2);
	if (nas == NERR_Success) {
		wcscpy(wComment, lgi2->lgrpi1_comment);
		NetApiBufferFree(lgi2);
		return NERR_Success;
	}
	else  {
		fprintf(stderr,"Unable to retrieve group information for\nGroup: %ls on server %ls\n",
			wGroupName, wComputerName);
		DisplayErrorText( nas );
		if (otfi != stdout) {
			fprintf(otfi,"Unable to retrieve group information for\nGroup: %ls on server %ls\n",
				wGroupName, wComputerName);
			fprintf(otfi, "%s\n", ErrorMessage);
		}
		if (otfi == stdout)
			return RTN_ERROR;
		fclose(otfi);
		strcpy(cmd,"NOTEPAD.EXE ");
		strcat(cmd,OutFileName);
		WinExec(cmd,SW_SHOW);
		return RTN_ERROR;
	}
}



GetFullNameFromDomain(LPWSTR lgrmi2_domainandname, char *FullName, FILE *otfi, 
					  char **Domn, LPWSTR *wPDC, int *DomainKt, int *DomainMax, char dlim) {
	LPWSTR wdelim;
	char temp[_MAX_PATH];
	LPWSTR wPDCName;
	int dom;
	DWORD nas;

	// Parse out domain name from user name
	wdelim = wcschr(lgrmi2_domainandname,L'\\');
	sprintf(temp,"%ls",lgrmi2_domainandname);
	*wdelim = 0;

	sprintf(temp,"%ls",lgrmi2_domainandname);
								
	// Look up PDC name in list
	for (dom=0,wPDCName=NULL; dom < *DomainKt && wPDCName==NULL; ++dom) {
		if (stricmp(Domn[dom],temp) == 0)
			wPDCName = wPDC[dom];
	}
	
	if (wPDCName == NULL) {
		// PDCName not found; look up on network and add to list
		nas = NetGetDCName(NULL,lgrmi2_domainandname,(LPBYTE *)&wPDCName);
		if (nas == NERR_Success) {
			add_new_domain(Domn, wPDC, DomainKt, DomainMax, lgrmi2_domainandname, wPDCName);
			*wdelim = L'\\';
		}	
		else {
			fprintf(stderr,"Unable to locate domain controller for %ls\n", lgrmi2_domainandname);
			DisplayErrorText( nas );
			fprintf(otfi,"%s\n",ErrorMessage);
			if (otfi != stdout)
				fclose(otfi);
			wPDCName = NULL;
		}
	}

	// get info about user from the pdc
	if (wPDCName != NULL) {
		// Have a PDC, look up User Name
		GetFullName(wPDCName, wdelim+1, FullName,dlim ? 254 : FULLNAMELEN);
	}
}


parse_Name(LPWSTR lgrmi2_domainandname, char *UserName, char *DomainName) {
	// extract domain name and user name
	char *temp, *delim;
	int l;
	l = wcslen(lgrmi2_domainandname);
	temp = malloc(l+5);
	sprintf(temp,"%ls",lgrmi2_domainandname);
//	if (strlen(temp) > 31) {
//		temp[31] = 0;
//		strncpy(UserName,temp,31);
//	}
//	else 
//		strcpy(UserName,temp);
	delim = strchr(temp,'\\');
	if (delim == NULL)
		*DomainName = 0;
	else {
		*delim = 0;
		if (strlen(temp) > 31)
			temp[31] = 0;
		strcpy(DomainName,temp);
		if (strlen(delim+1) > 31)
			*(delim+32) = 0;
		strcpy(UserName,delim+1);
	}
						
	strupr(DomainName);
}

int display_group_member(LPWSTR lgrmi2_domainandname, SID_NAME_USE lgrmi2_sidusage, 
	 PSID lgrmi2_sid, char *FullName, char fullname, char verbose, char terse, 
	 char groupmembers, FILE *otfi, LPWSTR wComment, int *DomainKt, int *DomainMax, 
	 char **Domn, LPWSTR *wPDC, char delim, char *GroupName, char *ServerName) {

	char DomainName[33], UserName[256], temp[256];
	DWORD SidLen = sizeof(temp);
	int rc = 0, ii=0, jj=0;

	// Get DomainName and UserName
	parse_Name(lgrmi2_domainandname,UserName, DomainName);

	// Display Name data based on usage
	switch(lgrmi2_sidusage) {
	case SidTypeUser:
		if (fullname) {
			GetFullNameFromDomain(lgrmi2_domainandname, FullName, otfi,
				Domn, wPDC, DomainKt, DomainMax, delim);
			// attempt to split at commas if possible
			if (!delim) 
				ii = splitname(FullName,FULLNAMELEN);
		}
		else
			*FullName = 0;

		if (verbose) {
			strcat(DomainName,"\\");
			if (delim) 
				delimPrint(ServerName,GroupName,"User", DomainName,"",UserName,FullName,0);
			else {
				if (strlen(UserName) > 30)
					UserName[30] = 0;
				strcat(DomainName,UserName);
				if (fullname) {
					sprintf(temp,"User:  %-38s %-38s\n",DomainName,Fname[0]);
					fprintf(otfi,temp);
					for (jj=1; jj < ii; ++jj) {
						sprintf(temp,"       %-38s %-38s\n"," ",Fname[jj]);
						fprintf(otfi,temp);
					}
				}
				else {
					sprintf(temp,"User:  %s\n", DomainName);
					fprintf(otfi,temp);
				}
			}
			break;
		}
		// Not Verbose

		if (delim) {
			delimPrint(ServerName,GroupName,"User", DomainName,"",UserName,FullName,0);
			break;
		}

		// set username to max of 20 characters
		if (strlen(UserName) > 20)
			UserName[20] = 0;
		sprintf(temp,"%-20s %-38s %s\n", UserName, Fname[0], DomainName);
		fprintf(otfi,temp);
		for (jj=1;jj < ii; ++jj) {
			sprintf(temp,"%-20s %-38s\n"," ",Fname[jj]);
			fprintf(otfi,temp);
		}

		break;
	case SidTypeUnknown:
		if (strlen(UserName) == 0) {
			GetTextualSid(lgrmi2_sid, temp, &SidLen);
			strcpy(UserName,"[Unknown]");
		}
		else
			strcpy(temp,"[Unknown]");
		if (delim) {
			delimPrint(ServerName,GroupName,"Unknown",DomainName,"",temp,"",0);
			break;
		}
		if (verbose) {
			strcat(DomainName,"\\");
			strcat(DomainName,UserName);
			fprintf(otfi,"Unknown: %-38s %s\n", DomainName, temp);
		}
		else 
			fprintf(otfi,"%-20s %-38s %-17s %s\n", UserName, "", DomainName, temp);
		break;

	case SidTypeDeletedAccount:
		if (strlen(UserName) == 0) {
			GetTextualSid(lgrmi2_sid, temp, &SidLen);
			strcpy(UserName,"[Deleted]");
		}
		else
			strcpy(temp,"Deleted");
		if (delim) {
			delimPrint(ServerName,GroupName,"Deleted",DomainName,"",UserName,temp,0);
			break;
		}

		if (verbose) {
			strcat(DomainName,"\\");
			strcat(DomainName,UserName);
			fprintf(otfi,"Deleted: %-38s %s\n", DomainName, temp);
		}
		else
			fprintf(otfi,"%-20s %-38s %-17s %s\n", UserName,"", DomainName,temp);
		break;

	case SidTypeWellKnownGroup:
		if (delim) {
			delimPrint(ServerName,GroupName,"Group",DomainName,UserName,"","",0);
			break;
		}
		if (verbose) {
			strcat(DomainName,"\\");
			strcat(DomainName,UserName);
			fprintf(otfi,"Group: %s\\%s\n",strupr(DomainName), UserName);
		}
		else 
			fprintf(otfi,"%-20s %-38s %-17s\n", "*Member Group:", UserName, strupr(DomainName));
		break;

	default: // Global Group
		*wComment = 0;
		if (verbose) {
			strupr(DomainName);
			if (!delim) {
				strcat(DomainName,"\\");
				strcat(DomainName,UserName);
				fprintf(otfi,"Group: %-38s %38s\n",DomainName, " ");
			}
		}
		else {
			if (!delim)
				fprintf(otfi,"%-20s %-38s %-17s\n", "*Member Group:", UserName, strupr(DomainName));
		}
		if (terse | groupmembers) {
			if (delim)
				delimPrint(ServerName,GroupName,"Group",DomainName,UserName,"","",0);
//			else
//				fprintf(otfi,"\n");
		}
		else // Call ListGlobalGroupMembers -- this will also print group comment
			if (ListGlobalGroupMembers(lgrmi2_domainandname, DomainKt, DomainMax, 
				Domn, wPDC, otfi, delim, GroupName, ServerName, groupmembers))
				rc = RTN_ERROR;
		break;
	}
	return rc;
}

int list_groups(int starting_member, char list, char members, char groupmembers,
				char fullname, char terse, char verbose, LPWSTR wComputerName, 
				LPWSTR wGroupName, FILE *otfi, char *OutFileName, char **Domn, 
				LPWSTR *wPDC, int *DomainKt, int *DomainMax, LPWSTR wDomNm, 
				char delim, char *excludes) {

	int rc = 0;
	WCHAR		wComment[MAXCOMMENTSZ+1];
	   

	LOCALGROUP_MEMBERS_INFO_2 *pgm2;
	LOCALGROUP_INFO_1 *lgi1,  LocalGroupInfo1[1];

	DWORD entries_read, total_entries, resume_handle, localgroup;
	DWORD entriesread, totalentries, resumehandle;

    char GroupName[256], groups[320], ComputerName[64];
	
   unsigned i,localgrouplen=GBUFSIZ;

   NET_API_STATUS  nas;

   char cmd[_MAX_PATH],temp[2048];
   pgm2 = NULL;
   lgi1 = NULL;


	// If extracting a list of members, run enum to get an array of groups
	resumehandle=0;
	entriesread=0;
	totalentries=1;
	
	sprintf(temp,"%ls",wComputerName);
	strcpy(ComputerName,temp);
	sprintf(temp,"%ls",wGroupName);
	strcpy(GroupName,temp);

	if (delim) {
		if (verbose)
			if (groupmembers)
				fprintf(otfi,"LocalGroupName,MemberType,MemberName,MemberFullName,MemberGroup\n");
			else
				fprintf(otfi,"LocalGroupName,MemberType,MemberName,MemberFullName\n");
		else
			fprintf(otfi,"ServerName,LocalGroupName,MemberType,DomainName,MemberGroupName,MemberUserName,MemberFullName\n");
	}
	strupr(excludes);
	strcat(excludes,",");

	do {
		// Why, you ask is the Enum being done inside the loop
		// Enum will sometimes return a partial list
		// Therefor, if list, we enum, otherwise fake it with a
		// single entry.  If Enum_Groups returns an error, quit this
		// routine and try again with the next computer or domain in the
		// list
		nas = Enum_Groups(list, wComputerName, wGroupName, &lgi1, LocalGroupInfo1, 
			&entriesread, &totalentries, &resumehandle, wComment, otfi);
		if (nas == RTN_ERROR) // A call to NetLocalGroupEnum failed, done for now
			return RTN_ERROR;

		for (localgroup=0; localgroup < entriesread; ++localgroup) {
		// LPWSTR   lgrpi1_comment;
			sprintf(GroupName,"%ls",lgi1[localgroup].lgrpi1_name);
			strcpy(temp,",");
			strcat(temp,GroupName);
			strcat(temp,",");
			strupr(temp);
			if (strstr(excludes,temp) != NULL)
				continue;
			wGroupName = lgi1[localgroup].lgrpi1_name;
			if (list) 
				wcscpy(wComment, lgi1[localgroup].lgrpi1_comment);
			else {
				// Get the comment, if the call fails exit this routine and try next computer
				if (GetLocalGroupComment(wComputerName, wGroupName, wComment, 
					otfi, OutFileName) == RTN_ERROR)
					return RTN_ERROR;
			}
			
			if (delim) {
				if (verbose)
					fprintf(otfi,"%ls\\%ls,GroupName,,\"%ls\"\n",wDomNm,wGroupName,wComment);
				else
					fprintf(otfi,"%ls,%ls,GroupName,,,,\"%ls\"\n",wDomNm,wGroupName,wComment);
			}
			else {
				if (terse) {
					if (verbose) 
						if (fullname) {
							sprintf(temp,"%ls\\%ls",wDomNm,wGroupName);
							fprintf(otfi,"%-59s\t%ls\n",temp,wComment);
						}
						else
							fprintf(otfi,"%ls\\%ls\n",wDomNm,wGroupName);
					else
						if (fullname)
							fprintf(otfi,"%-38ls%ls\t%ls\n",wGroupName,wDomNm,wComment);
						else
							fprintf(otfi,"%-38ls%ls\n",wGroupName,wDomNm);
				}
			}
			if (list & terse)
				continue;
			
			sprintf(temp,"%ls\\%ls",wDomNm, wGroupName);
			if (!fullname)
				*wComment = 0;

			if (localgroup && members && !delim) // Not the first group, add a blank line
				fprintf(otfi,"\n");
			if (!delim) {
				if (verbose)
					fprintf(otfi,"*Local %-76s\n", temp); 
				else
					fprintf(otfi,"%-20s %-38ls %-17ls\n","*Local Group:", wGroupName, wcsupr(wDomNm));
				if (*wComment)
					fprintf(otfi," %ls\n",wComment);
			}

		// Get the type 2 info into pgm2

			entries_read = 0;
			total_entries= 1;
			*groups = 0;
				resume_handle = 0;
			// Read in all of the Local Groups
			// First read in without indirect flag
			while (entries_read < total_entries && members) {
				nas = NetLocalGroupGetMembers(wComputerName,wGroupName,2,(LPBYTE *) &pgm2,
					GBUFSIZ,&entries_read, &total_entries, &resume_handle);

				if (nas == NERR_Success || nas== ERROR_MORE_DATA) {
					// Found More
					for (i=0; i < entries_read; ++i) {
						// Display the Name based on the SID USage
						sprintf(temp,"%ls",pgm2[i].lgrmi2_domainandname);
						rc = display_group_member(pgm2[i].lgrmi2_domainandname,
							pgm2[i].lgrmi2_sidusage, pgm2[i].lgrmi2_sid,FullName, fullname, verbose, 
							terse, groupmembers, otfi, wComment, DomainKt, 
							DomainMax, Domn, wPDC, delim, GroupName, ComputerName);
					}
					if (entries_read == 0)
						if (delim)
							delimPrint(ComputerName,GroupName,"EmptyGroup","","","","",0);
				}
				else  {
					fprintf(stderr,"Unable to retrieve group users for\nGroup: %ls in %ls\n",
						wGroupName, wComputerName);
					DisplayErrorText( nas );
					fprintf(otfi,"Unable to retrieve group users for\nGroup: %ls in %ls\n",
						wGroupName, wComputerName);
					fprintf(otfi, "%s\n", ErrorMessage);
					if (otfi == stdout)
						return RTN_ERROR;
					fclose(otfi);
					strcpy(cmd,"NOTEPAD.EXE ");
					strcat(cmd,OutFileName);
					WinExec(cmd,SW_SHOW);
					return RTN_ERROR;
				}

			}	
		}
	} while (resumehandle);
	if (pgm2)
		NetApiBufferFree((PVOID) pgm2);
	return rc;
		
}



int __cdecl wmain(int argc, wchar_t *argv[] ) {

	int stdio_handle, rc = 0;
	int starting_member = 0;
	char add = 0, del = 0, set = 0, delim = 0, addgroup = 0, delgroup = 0;
    LPWSTR      wComputerName = NULL; // default to local machine
	LPWSTR		wGroupName;
	WCHAR		wComments[1024];
	WCHAR		wDomNm[64];
	LPWSTR		*wCName;
	char **Domn;
	LPWSTR *wPDC;
	char excludes[8192];
	
	int DomainKt = 0, DomainMax = 255;	   

//	LOCALGROUP_MEMBERS_INFO_2 *pgm2;
	LOCALGROUP_MEMBERS_INFO_3 localgroup_members[USER_COUNT];
//	LOCALGROUP_INFO_1 *lgi1;

	int i,count;

    char ComputerName[32];
    char GroupName[256], groups[320], terse=0;
    char members=1, list = 0, groupmembers=0;
    int CNameArraySize=64, CNameCount=0;
	
   unsigned localgrouplen=GBUFSIZ;
   int currentname;

//   NET_API_STATUS  nas;
	char tmp[_MAX_PATH],*p,cmd[_MAX_PATH], OutFilePath[_MAX_PATH];

	otfi = stdout;
   groups[0] = 0;
   *wComments = 0;
   // Notepad will be used to either display the data or the help
   // Get the path to a temp directory here
	if ((p=getenv("tmp")) == NULL)
		if ((p = getenv("temp")) == NULL)
			GetWindowsDirectory(tmp,_MAX_PATH);
		else
			strcpy(tmp,p);
	else
		strcpy(tmp,p);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");
   
	if( argc < 3 ) {
	   help(argv,tmp);
       return RTN_USAGE;
	}
 

	// NetApiBufferAllocate(GBUFSIZ, &gi0);

 //   localgroups = (char *)malloc(GBUFSIZ);
 // *localgroups = 0;
	
    wGroupName = argv[1];
    wComputerName = argv[2];
	
	wCName = malloc(CNameArraySize * sizeof(LPWSTR));
	Domn = malloc(sizeof(char *) * DomainMax);
	wPDC = malloc(sizeof(LPWSTR) * DomainMax);

	if (wCName == NULL || Domn == NULL || wPDC == NULL) {
		fprintf(stderr, "Insufficient Memory!\n");
		return 4;
	}


	sprintf(ComputerName,"%ls",wComputerName);
	sprintf(GroupName,"%ls", wGroupName);


	stdio_handle = _fileno(stdout);
	// Process Command Line Arguments
	starting_member = argc+1;
	set_flags(argc, argv, &members, &verbose, &fullname, &terse, &groupmembers, 
		&starting_member, &add, &del, &set, &slow_but_steady, &delim, &addgroup, 
		&delgroup, wComments,excludes);
	
	// Gather list of computer names / domain names and store in wCName
	wCName = gather_computer_names(argv, starting_member, &CNameCount, &CNameArraySize, otfi);

	// Looking at a single group or all groups?
	if (stricmp(GroupName,"-LIST") == 0 || strcmp(GroupName,"*") == 0) {
		list = 1;
//		members = 0;
		if (add || del || set || addgroup || delgroup || *wComments) {
			fprintf(stderr,"Group Name Required.\n");
			return 4;
		}
	}

	// if adding or deleting, gather the list of members to add/ del
	// do this once, it's not going to change

	if (add || del || set) 
		gather_users(otfi, starting_member, argc, argv, localgroup_members, &count);

	// The following code must be repeated for each computer or domain specified in the array
	for (currentname = 0; currentname < CNameCount; ++currentname) {

		sprintf(ComputerName, "%ls", wCName[currentname]);
		wcscpy(wComputerName, wCName[currentname]);
		// Send output to either stdout or to a file depending on
		// whether input was redirected

		otfi = OpenOutfile(stdio_handle, GroupName, ComputerName, list, tmp, OutFilePath);

		// Look for standard abbreviations: -NE, ...
		expand_abbreviation(wDomNm, wComputerName);

		// Add to list of domains, if can't find PDC, skip this iteration
		if (add_to_domainlist(otfi, wDomNm, wCName[currentname], Domn, 
			wPDC, wComputerName, &DomainKt, &DomainMax) == RTN_ERROR)
			continue;

		sprintf(ComputerName,"%ls",wComputerName);

		// If adding members, adding groups or listing data

		if (add || del || set) 
			add_delete_users(otfi, wComputerName, wGroupName, add, del, set, localgroup_members, count);
		else
			if (addgroup || delgroup || *wComments)
				add_delete_groups(otfi, wComputerName, wGroupName, wComments, 
				(addgroup ? 'A' : (delgroup ? 'D' : 'C')));
			else 
				rc = list_groups(starting_member, list, members, groupmembers, 
				fullname, terse, verbose, wComputerName, wGroupName, otfi, OutFilePath, 
				Domn, wPDC, &DomainKt, &DomainMax, wDomNm, delim, excludes);
	
	
		if (otfi == stdout)
			fflush(otfi);
		else {
			fclose(otfi);
			strcpy(cmd,"NOTEPAD.EXE ");
			strcat(cmd,OutFilePath);
			WinExec(cmd,SW_SHOW);
		}
	
	}

//	NetApiBufferFree((PVOID) gi0);
	if (add || del) {
		for (i=0; i < count; ++i)
			free(localgroup_members[i].lgrmi3_domainandname);
	}
	else {
		for (i=0; i < DomainKt; ++i) {
			free (Domn[i]);
			free (wPDC[i]);
		}
		free (Domn);
		free (wPDC);
	}


	free((void *)wCName);
    return rc;
}
 
void DisplayErrorText(DWORD dwLastError) {
       HMODULE hModule = NULL; // default to system source
       LPSTR MessageBuffer;
       DWORD dwBufferLength;
       DWORD dwFormatFlags;
 
       dwFormatFlags = FORMAT_MESSAGE_ALLOCATE_BUFFER |
                       FORMAT_MESSAGE_IGNORE_INSERTS |
                       FORMAT_MESSAGE_FROM_SYSTEM ;
 
       //
       // if dwLastError is in the network range, load the message source
       //
       if(dwLastError >= NERR_BASE && dwLastError <= MAX_NERR) {
           hModule = LoadLibraryEx(
               TEXT("netmsg.dll"),
               NULL,
               LOAD_LIBRARY_AS_DATAFILE
               );
 
           if(hModule != NULL)
               dwFormatFlags |= FORMAT_MESSAGE_FROM_HMODULE;
       }
 
       //
       // call FormatMessage() to allow for message text to be acquired
       // from the system or the supplied module handle.
       //
       if(dwBufferLength = FormatMessageA(
           dwFormatFlags,
           hModule, // module to get message from (NULL == system)
           dwLastError,
           MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // default language
           (LPSTR) &MessageBuffer,
           0,
           NULL
           ))
       {
           DWORD dwBytesWritten;
 
           //
           // Output message string on stderr
           //
           WriteFile(
               GetStdHandle(STD_ERROR_HANDLE),
               MessageBuffer,
               dwBufferLength,
               &dwBytesWritten,
               NULL
               );
 
           //
           // free the buffer allocated by the system
           //
		   strcpy(ErrorMessage, MessageBuffer);
           LocalFree(MessageBuffer);
       }
 
       //
       // if you loaded a message source, unload it.
       //
       if(hModule != NULL)
           FreeLibrary(hModule);


}

